/*
 * Copyright (C) 2005-2012 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file filter.c The stand-alone filter program. */

#include <regex.h>

#include "para.h"
#include "filter.cmdline.h"
#include "list.h"
#include "sched.h"
#include "ggo.h"
#include "buffer_tree.h"
#include "filter.h"
#include "string.h"
#include "stdin.h"
#include "stdout.h"
#include "error.h"
#include "version.h"

/** The list of all status items used by para_{server,audiod,gui}. */
const char *status_item_list[] = {STATUS_ITEM_ARRAY};

char *stat_item_values[NUM_STAT_ITEMS] = {NULL};

/** Initialize the array of errors for para_filter. */
INIT_FILTER_ERRLISTS;

/** The task that reads from stdin. */
static struct stdin_task stdin_task_struct;
/** pointer to the stdin task. */
static struct stdin_task *sit = &stdin_task_struct;

/** The task that writes converted data to stdout. */
static struct stdout_task stdout_task_struct;
/** Pointer to the stdout task. */
static struct stdout_task *sot = &stdout_task_struct;

/** Gengetopt struct that holds the command line args. */
static struct filter_args_info conf;

static int loglevel;
INIT_STDERR_LOGGING(loglevel);

__noreturn static void print_help_and_die(void)
{
	int d = conf.detailed_help_given;
	const char **p = d? filter_args_info_detailed_help
		: filter_args_info_help;

	printf_or_die("%s\n\n", FILTER_CMDLINE_PARSER_PACKAGE "-"
		FILTER_CMDLINE_PARSER_VERSION);
	printf_or_die("%s\n\n", filter_args_info_usage);
	for (; *p; p++)
		printf_or_die("%s\n", *p);
	print_filter_helps(d);
	exit(0);
}

static int parse_config(int argc, char *argv[])
{
	static char *cf; /* config file */
	struct stat statbuf;

	if (filter_cmdline_parser(argc, argv, &conf))
		return -E_FILTER_SYNTAX;
	HANDLE_VERSION_FLAG("filter", conf);
	if (conf.help_given || conf.detailed_help_given)
		print_help_and_die();
	loglevel = get_loglevel_by_name(conf.loglevel_arg);
	if (!cf) {
		char *home = para_homedir();
		cf = make_message("%s/.paraslash/filter.conf", home);
		free(home);
	}
	if (!stat(cf, &statbuf)) {
		struct filter_cmdline_parser_params params = {
			.override = 0,
			.initialize = 0,
			.check_required = 0,
			.check_ambiguity = 0,
			.print_errors = 1
		};
		if (filter_cmdline_parser_config_file(cf, &conf, &params))
			return -E_FILTER_SYNTAX;
	}
	if (!conf.filter_given)
		return -E_NO_FILTERS;
	return 1;
}

/**
 * The main function of para_filter.
 *
 * Para_filter reads data from stdin, converts it by using a chain
 * of filters (specified on the command line) and writes the resulting
 * data to stdout.
 *
 * \param argc Number of command line options.
 * \param argv Vector of arguments.
 *
 * \return \a EXIT_SUCCESS on success, EXIT_FAILURE on errors.
 */
int main(int argc, char *argv[])
{
	static struct sched s;
	int i, ret;
	struct filter *f;
	struct btr_node *parent;
	struct filter_node **fns;

	filter_init();
	ret = parse_config(argc, argv);
	if (ret < 0)
		goto out;
	sit->btrn = btr_new_node(&(struct btr_node_description)
		EMBRACE(.name = "stdin"));
	stdin_set_defaults(sit);
	register_task(&s, &sit->task);

	fns = para_malloc(conf.filter_given * sizeof(*fns));
	for (i = 0, parent = sit->btrn; i < conf.filter_given; i++) {
		char *fa = conf.filter_arg[i];
		struct filter_node *fn;

		fn = fns[i] = para_calloc(sizeof(*fn));
		ret = check_filter_arg(fa, &fn->conf);
		if (ret < 0) {
			free(fn);
			goto out_cleanup;
		}
		fn->filter_num = ret;
		f = filters + fn->filter_num;
		sprintf(fn->task.status, "%s", f->name);
		PARA_DEBUG_LOG("filter #%d: %s\n", i, f->name);
		fn->btrn = btr_new_node(&(struct btr_node_description)
			EMBRACE(.name = f->name, .parent = parent,
			.handler = f->execute, .context = fn));
		fn->task.pre_select = f->pre_select;
		fn->task.post_select = f->post_select;
		f->open(fn);
		register_task(&s, &fn->task);
		parent = fn->btrn;
	}
	sot->btrn = btr_new_node(&(struct btr_node_description)
		EMBRACE(.name = "stdout", .parent = parent));
	stdout_set_defaults(sot);
	register_task(&s, &sot->task);

	s.default_timeout.tv_sec = 1;
	s.default_timeout.tv_usec = 0;
	btr_log_tree(sit->btrn, LL_INFO);
	ret = schedule(&s);
out_cleanup:
	for (i--; i >= 0; i--) {
		struct filter_node *fn = fns[i];

		f = filters + fn->filter_num;
		if (f->close)
			f->close(fn);
		btr_remove_node(&fn->btrn);
		free(fn->conf);
		free(fn);
	}
	free(fns);
	btr_remove_node(&sit->btrn);
	btr_remove_node(&sot->btrn);
out:
	if (ret < 0)
		PARA_EMERG_LOG("%s\n", para_strerror(-ret));
	exit(ret < 0? EXIT_FAILURE : EXIT_SUCCESS);
}
